#include <stdlib.h>
#include <string>

#include "l3ds.h"
#include "tga.h"


/***** Aide OpenGL *****/

/** Rapporte les erreurs OpenGL, s'il y en a, sur stderr */
void checkError(const char *functionName)
{
	GLenum error;
	while(( error = glGetError() ) != GL_NO_ERROR) {
		fprintf(stderr, "GL error 0x%X detected in %s\n", error, functionName);
	}
}

/** Affichage d'informations sur le hardware et OpenGL */
void dumpInfo(void)
{
	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	checkError("dumpInfo");
}

const GLvoid *bufferObjectPtr(GLsizei index)
{
	return(const GLvoid *)(((char *) NULL) + index);
}


/* vnements (souris + modifieurs) */
#define MOUSE_SHIFT  0x0001
#define MOUSE_CTRL   0x0002
#define MOUSE_ALT    0x0004
#define MOUSE_LEFT   0x0010
#define MOUSE_MIDDLE 0x0020
#define MOUSE_RIGHT  0x0040

struct CMouse
{
	double	X ,Y,            /* Position normalise     */
			X0,Y0,			 /* Position prcdente     */
			Dx,Dy;			 /* Dplacement             */
	int	Left, Middle, Right; /* Etat des boutons        */
	int	Shift, Ctrl, Alt;    /* Modifieurs              */
	int	State;               /* Etat global             */
};
// Flags pour les tats de la souris
#define SHIFT  0x0001
#define CTRL   0x0002
#define ALT    0x0004
#define LEFT   0x0010
#define MIDDLE 0x0020
#define RIGHT  0x0040

// Structure reprsentant une vue (camra, lampe...)
struct CView
{
	float	Distance;
	float	AngleY;
	float	AngleX;
};

/*********************************************************************************/

/***** Aide chargement maillage *****/

//Lemming mesh
//Vertices XYZW (vec4) float
//indices GL_UNSIGNED_SHORT
//Primitives: triangles


L3DS scene3DS;
LColor3 *colorArray;

GLuint lemmingVertexArrayObject;


void initModel(const char* fileName){
	scene3DS.LoadFile(fileName);
	scene3DS.GetMesh(0).NormalizeMesh();
	scene3DS.GetMesh(0).MapUVToSphereMap();
	scene3DS.GetMesh(0).Optimize(oFull);

	LMesh &mesh = scene3DS.GetMesh(0);

	colorArray=new LColor3[mesh.GetVertexCount()];
	for(unsigned int i=0; i<mesh.GetTriangleCount(); i++){
		LColor3 col=scene3DS.GetMaterial(mesh.GetTriangle2(i).materialId).GetDiffuseColor();
		colorArray[mesh.GetTriangle(i).a]=col;
		colorArray[mesh.GetTriangle(i).b]=col;
		colorArray[mesh.GetTriangle(i).c]=col;		
	}
}

GLfloat *getModelVertices(int m=0){
	LMesh &mesh = scene3DS.GetMesh(0);
	return (GLfloat *)&mesh.GetVertex(0);
}

GLfloat *getModelNormals(int m=0){
	LMesh &mesh = scene3DS.GetMesh(0);
	return (GLfloat *)&mesh.GetNormal(0);
}
GLfloat *getModelColors(int m=0){
	return (GLfloat*)colorArray;
}

GLfloat *getModelUV(int m=0) {
	LMesh &mesh = scene3DS.GetMesh(0);
	return (GLfloat*)&mesh.GetUV(0);
}

GLfloat *getModelTangents(int m=0) {
	LMesh &mesh = scene3DS.GetMesh(0);
	return (GLfloat*)&mesh.GetTangent(0);
}

GLfloat *getModelBinormals(int m=0) {
	LMesh &mesh = scene3DS.GetMesh(0);
	return (GLfloat*)&mesh.GetBinormal(0);
}

unsigned short *getModelIndices(int m=0){
	LMesh &mesh = scene3DS.GetMesh(0);
	return (unsigned short *)&mesh.GetTriangle(0);
}

int getModelNumIndices(int m=0){
	LMesh &mesh = scene3DS.GetMesh(0);
	return mesh.GetTriangleCount()*3;
}

void initMaillageAttrib(GLuint shaderProgramID){

	//Vertices positions
	//Get attribute ID
	GLuint attribID=glGetAttribLocation( shaderProgramID, "VertexPosition" );
	if(attribID != -1){
		//Connect data array to attribute
		glVertexAttribPointer( attribID, 4, GL_FLOAT, GL_FALSE, 0, getModelVertices(0));
		//Attribute array will be used for rendering
		glEnableVertexAttribArray( attribID );
	}

	//Vertices normals
	attribID=glGetAttribLocation( shaderProgramID, "VertexNormal" );
	if(attribID != -1){
		glVertexAttribPointer( attribID, 3, GL_FLOAT, GL_FALSE, 0, getModelNormals(0));
		glEnableVertexAttribArray( attribID );
	}

	//Vertices colors
	attribID=glGetAttribLocation( shaderProgramID, "VertexColor" );
	if(attribID != -1){
		glVertexAttribPointer( attribID, 3, GL_FLOAT, GL_FALSE, 0, getModelColors(0));
		glEnableVertexAttribArray( attribID );
	}

	//Vertices UV
	attribID=glGetAttribLocation( shaderProgramID, "VertexUV" );
	if(attribID != -1){
		glVertexAttribPointer( attribID, 2, GL_FLOAT, GL_FALSE, 0, getModelUV(0));
		glEnableVertexAttribArray( attribID );
	}

	//Vertices Tangents
	attribID=glGetAttribLocation( shaderProgramID, "VertexTangent" );
	if(attribID != -1){
		glVertexAttribPointer( attribID, 3, GL_FLOAT, GL_FALSE, 0, getModelTangents(0));
		glEnableVertexAttribArray( attribID );
	}

	//Vertices Binormal
	attribID=glGetAttribLocation( shaderProgramID, "VertexBinormal" );
	if(attribID != -1){
		glVertexAttribPointer( attribID, 3, GL_FLOAT, GL_FALSE, 0, getModelBinormals(0));
		glEnableVertexAttribArray( attribID );
	}

	checkError("initMaillage");
}


/*******************************************************************************************/

/**** Aide chargement textures ****/

void initTexture(const char* fileName, GLuint *id) {
	//Initialisation et chargement d'une texture 2D

	const GLenum texMode = GL_TEXTURE_2D;

	glGenTextures(1, id);	
	glBindTexture(texMode, *id);

	glTexParameteri(texMode, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(texMode, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	
	glTexParameteri(texMode, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(texMode, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	if (load_tga_texture(fileName,texMode) == 0)
		fprintf(stderr,"Error while loading texture");

	checkError("initTexture");
}